/**
  General-Purpose Graphics Processing Unit (GPGPU)
  Demo Application for neural networks / matrix multiply.
  
  Neural Network basics (sigmoid, layer counts) are courtesy of 
  Seth Snedigar's 2008 CS641 Project2 code.  Basic idea (GPU-NN,
  packing channels, many-problem evaluation) is by Jay Schamel circa 2006.

  Blondie24 layer sizes: 32, 40, 10, 1.

  Tricky part: to make glReadPixels run efficiently, you've got to fetch 10K+ pixels
  at a time.  Similarly, each framebuffer object bind will cost you 15-30us, which 
  is tens of thousands of pixels.  So you've *got* to run thousands of evaluations
  at a time to have any hope of getting decent efficiency on the graphics card.

  Orion Sky Lawlor, olawlor@acm.org, 2008-04-24 (Public Domain)
*/
#include "ogl/gpgpu.h" /* Orion's GPGPU library */
#include "ogl/glsl.cpp" /* for easy linking, just #include the library code */

/* Exit with error */
void die(const char *why) {fprintf(stderr,"FATAL ERROR> %s\n",why);exit(1);}

/** Given an integer, return a string */
std::string str(int i) {char buf[100]; sprintf(buf,"%d",i); return buf;}
std::string str(float f) {char buf[100]; sprintf(buf,"%.8f",f); return buf;}

/** Choose between two strings based on a boolean */
std::string if_then_else(bool condition,const std::string &then_str,const std::string &else_str) {
	if (condition) return then_str;
	else return else_str;
}

/* Standard neural network sigmoid function. */
float sigmoid(float x) {
	return 2/(1 + exp(-x)) - 1;
}

/** 
 Class to evaluate a single layer of a neural network 
 on many different inputs, using the graphics card. 
*/
class gpgpu_NN_layer_evaluator {
public:
	/* Create an evaluator that computes, for each problem,
	 our network's output node as a function of its input nodes.
	
	Conceptually, we calculate:
	
	for (int o=0;o<outputs;o++) {
		for (int p=0;p<problems;p++) {
			float activation=0;
			for (int i=0;i<inputs;i++) {
				activation+=inputs(p,i)*weights(i,o);
			}
			outputs(p,o)=sigmoid(activation);
		}
	}
	
	input and output *problem* values are packed into the *colors*
	of each pixel.  E.g., problem 0 is red, problem 1 is green,
	problem 2 is blue, problem 3 is alpha; problem 4 is red of next pixel, ...
	Weights of adjacent inputs are also packed into colors analogously.
	
	So we really compute:
	
	for (int o=0;o<outputs;o++) {
		for (int p=0;p<problems;p+=4) {
			vec4 activation=0;
			for (int i=0;i<inputs;i+=4) {
				activation+=inputs(p,i+0)*weights(i/4,o).r;
				activation+=inputs(p,i+1)*weights(i/4,o).g;
				activation+=inputs(p,i+2)*weights(i/4,o).b;
				activation+=inputs(p,i+3)*weights(i/4,o).a;
			}
			outputs(p,o)=sigmoid(activation);
		}
	}
	
	This means the number of problems *must* be a multiple of 4.
	*/
	gpgpu_NN_layer_evaluator(int problems,
		int inputs,gpgpu_texture2D *input,
		bool base_layer,gpgpu_texture2D *weight,
		int outputs);
	
	/* Once constructed, these variables hold our output values. 
	  Recall that outputs(o,p) holds a color-packed set of answers.
	*/
	gpgpu_texture2D *output;
	gpgpu_framebuffer *output_fb;
	
	/* Evaluate our outputs as a function of our inputs. */
	void run(void) {
		gpgpu_add(prog,input,"input",0);
		if (weight!=0)
			gpgpu_add(prog,weight,"weight",1);
		output_fb->run(prog);
	}
	
	gpgpu_texture2D *input, *weight; /* textures for input and weight data */
private:
	GLhandleARB prog; /* GLSL program to compute outputs from inputs */
};

gpgpu_NN_layer_evaluator::gpgpu_NN_layer_evaluator(int problems,
		int inputs,gpgpu_texture2D *input_,
		bool base_layer,gpgpu_texture2D *weight_,
		int outputs)
		:input(input_), weight(weight_)
{
	if (problems%4 != 0) die("Number of problems must be a multiple of 4!");
	if (inputs != input->height || problems/4 != input->width) 
		die("Input data size mismatch!");
	if ((!base_layer) && (outputs != weight->height || inputs/4 != weight->width)) 
		die("Weight array size mismatch!");
	output=new gpgpu_texture2D(problems/4,outputs);
	
	prog=makeProgramObject(GPGPU_V,(
		"varying vec2 location; /* onscreen tex coords: x==output, y==problem */ \n"
		"uniform vec2 locationScale; /* to onscreen pixels */ \n"
		"uniform sampler2D input; uniform vec2 inputScale; \n"
		"uniform sampler2D weight; uniform vec2 weightScale; \n"
		"vec4 sigmoid(vec4 x) {\n"
		"	return 2/(1 + exp(-x)) - 1;\n"
		"}\n"
		"void main(void) { /* implicitly loops over all our pixels--over output and problem number */\n"
	+if_then_else(base_layer, /* base layer special case: no weights, just copy over with sigmoid */
	/*then*/"	vec4 activation=texture2D(input,location);\n"
	,
	/*else*/"	vec4 activation=vec4(0.0);\n"
		"	for (int i=0;i<"+str(inputs/4)+";i++) {\n"
		"		vec4 w = texture2D(weight,vec2((i+0.5)*weightScale.x,location.y));\n"
		"		activation += texture2D(input,vec2(location.x,(4*i+0.5)*inputScale.y))*w.r;\n"
		"		activation += texture2D(input,vec2(location.x,(4*i+1.5)*inputScale.y))*w.g;\n"
		"		activation += texture2D(input,vec2(location.x,(4*i+2.5)*inputScale.y))*w.b;\n"
		"		activation += texture2D(input,vec2(location.x,(4*i+3.5)*inputScale.y))*w.a;\n"
		"	}\n"
	)+
		"	gl_FragColor=sigmoid(activation);\n"
		"}\n"
		).c_str()
	);
	output_fb=new gpgpu_framebuffer(output);
}

/* Return a random floating-point number between -1 and +1 */
float rand_float(void) {
	return (rand()&0xffff)*(2.0/0xffff)-1.0;
}

/** Return a texture full of neural network weight factors */
gpgpu_texture2D *make_weights(unsigned int outputs,unsigned int inputs) {
	float *weights=new float[inputs*outputs];
	for (unsigned int o=0;o<outputs;o++) {
		for (unsigned int i=0;i<inputs;i++) {
			weights[o*inputs + (i/4)*4 + (i%4)]=rand_float();
		}
	}
	gpgpu_texture2D *ret=new gpgpu_texture2D(inputs/4,outputs,weights);
	delete[] weights;
	return ret;
}

/** Return a set of boards to evaluate */
void get_boards(unsigned int nboards,unsigned int squares,float *dest) {
	for (unsigned int s=0;s<squares;s++) {
		for (unsigned int b=0;b<nboards;b++) {
			dest[s*nboards + (b/4)*4 + (b%4)]=rand_float();
		}
	}
}


/* Evaluate "problems" boards at a time (MUST be a multiple of 4) */
void time_problems(int problems)
{
	enum {nLayers=4};
	gpgpu_NN_layer_evaluator *eval[nLayers];
	int squares=32; /* checkerboard squares */
	int layer1=40; /* output nodes of first layer */
	int layer2=12; /* output nodes of second layer */
	int layer3=1; /* output nodes of final layer */
	
	/* Overall input data is contents of each square */
	gpgpu_texture2D *board=new gpgpu_texture2D(problems/4,squares);
	float *boardCPU=new float[4*(problems/4)*squares];
	float *evaluation=new float[problems];
	
	/* Input layer: no weights, output size same as input, n squares */
	eval[0]=new gpgpu_NN_layer_evaluator(problems, squares,board, true,0, squares);
	
	/* First layer: squares inputs, layer1 outputs */
	eval[1]=new gpgpu_NN_layer_evaluator(problems, squares,eval[0]->output, 
			false, make_weights(layer1,squares), layer1);
	
	/* Second layer: layer1 inputs, layer2 outputs */
	eval[2]=new gpgpu_NN_layer_evaluator(problems, layer1,eval[1]->output, 
			false, make_weights(layer2,layer1), layer2);
	
	/* Final layer: layer2 inputs, layer3 outputs */
	eval[3]=new gpgpu_NN_layer_evaluator(problems, layer2,eval[2]->output, 
			false, make_weights(layer3,layer2), layer3);
	
	/* Do some board evaluations */
	double t_start=gpgpu_timer();
	int repeat;
	for (repeat=0;repeat<100;repeat++) {
		/* Copy the input data onto the graphics card */
		get_boards(problems,squares,boardCPU);
		board->update_data(boardCPU);
		
		/* Push the input data through each layer of the network */
		for (int l=0;l<nLayers;l++) {
			eval[l]->run();
			if (0) { // Show & tell mode
				float pixel[4];
				eval[l]->output_fb->read(pixel,1,1);
				for (int p=0;p<4;p++) printf("	%.4g",pixel[p]);
				printf("\n");
				gpgpu_show_and_pause(eval[l]->input,("Layer "+str(l)+" input").c_str());
				if (l>0) gpgpu_show_and_pause(eval[l]->weight,("Layer "+str(l)+" weight").c_str());
				gpgpu_show_and_pause(eval[l]->output,("Layer "+str(l)+" output").c_str());
			}
		}
		
		eval[3]->output_fb->read(evaluation,problems/4,1);
		/* evaluation array now contains the neural network's evaluation of each board */
		if (0) { /* print results */
			for (int p=0;p<problems;p++)
				printf("	%.2g",evaluation[p]);
			printf("\n");
		}
	}
	double elapsed=(gpgpu_timer()-t_start)/repeat/problems;
	printf("%d problem-sized batch: %.3f ns per board; %.0f evaluations/second\n",
		problems, elapsed*1.0e9, 1.0/elapsed);
	
	// Dispose of all computation arrays
	for (int l=0;l<nLayers;l++) {
		delete eval[l]->output_fb;
		delete eval[l]->output;
		if (l>0) delete eval[l]->weight;
		delete eval[l];
	}
	delete[] evaluation;
	delete[] boardCPU;
	delete board;
}

int main(int argc,char *argv[]) {
	gpgpu_init();
	
	/* Vary the number of boards to evaluate in each pass */
	for (int problems=4;problems<=8192;problems*=2) 
		time_problems(problems);

	return 0;
}
